#include <channel.h>
#include <misc.h>                       /* get_param */
#include <stdlib.h>
#include <process.h>
#include <time.h>
#include <stdio.h>
#include <string.h>

#define DIM  3   /* Dimension of the Hypercube */
#define STRINGLENGTH 80
#define BUFFERSIZE ((1 << DIM)*STRINGLENGTH)

#ifdef ROOT

  /*{{{  PRINT*/
  #define PRINT1(arg) printf(arg)
  #define PRINT2(arg1,arg2) printf(arg1,arg2)
  #define PRINT3(arg1,arg2,arg3) printf(arg1,arg2,arg3)
  #define PRINT4(arg1,arg2,arg3,arg4) printf(arg1,arg2,arg3,arg4)
  #define PRINT5(arg1,arg2,arg3,arg4,arg5) printf(arg1,arg2,arg3,arg4,arg5)
  /*}}}  */

  /*{{{  GPRINT*/
  #define GPRINT1(arg) gprint1(arg)
  #define GPRINT2(arg1,arg2) printf(arg1,arg2)
  #define GPRINT3(arg1,arg2,arg3) printf(arg1,arg2,arg3)
  #define GPRINT4(arg1,arg2,arg3,arg4) printf(arg1,arg2,arg3,arg4)
  #define GPRINT5(arg1,arg2,arg3,arg4,arg5) printf(arg1,arg2,arg3,arg4,arg5)
  /*}}}  */

  /*{{{  INITIALIZE*/
  #define INITIALIZE  \
    set_abort_action(ABORT_HALT); \
    dim = *((int *) get_param(5)); \
    d_in = CreateChannels(get_param(3), dim); \
    d_out = CreateChannels(get_param(4), dim); \
    name = (int *) get_param(6); \
    level = node_level(*name); \
    bit = dim; \
    num_processors = (1 << dim)
  /*}}}  */

  #define GETC(f) getc(f)

  /*{{{  RAND*/
  #define RAND(x) \
     x = rand(); \
     {int j; \
        for (j = 0; j < dim; j++) \
          ChanOutInt(d_out[j], x);}
  /*}}}  */

  #define EXIT exit_terminate(0);

#else

  /*{{{  PRINT*/
  #define PRINT1(arg)
  #define PRINT2(arg1,arg2)
  #define PRINT3(arg1,arg2,arg3)
  #define PRINT4(arg1,arg2,arg3,arg4)
  #define PRINT5(arg1,arg2,arg3,arg4,arg5)
  /*}}}  */

  /*{{{  GPRINT*/
  #define GPRINT1(arg) gprint1(arg)
  #define GPRINT2(arg1,arg2) sprintf(INPUT,arg1,arg2)
  #define GPRINT3(arg1,arg2,arg3) sprintf(INPUT,arg1,arg2,arg3)
  #define GPRINT4(arg1,arg2,arg3,arg4) sprintf(INPUT,arg1,arg2,arg3,arg4)
  #define GPRINT5(arg1,arg2,arg3,arg4,arg5) sprintf(INPUT,arg1,arg2,arg3,arg4,arg5)
  /*}}}  */

  /*{{{  INITIALIZE*/
  #define INITIALIZE  \
    set_abort_action(ABORT_HALT); \
    dim = *((int *) get_param(2)); \
    d_in = CreateChannels(get_param(3), dim); \
    d_out = CreateChannels(get_param(4), dim); \
    name = (int *) get_param(1); \
    level = node_level(*name); \
    bit = dim; \
    num_processors = (1 << dim)
  /*}}}  */

  #define GETC(f) 0

  /*{{{  RAND*/
  #define RAND(x) \
    x = ChanInInt(d_in[level - 1]); \
    {int j; \
     for (j = level; j < dim; j++) \
       ChanOutInt(d_out[j], x);}
  /*}}}  */

  #define EXIT

#endif

/* common to both root and node */

/*{{{  PARREDUCE(f1, f2, r, s)*/
#define PARREDUCE(f1, f2, r, s) \
  {\
    num_processors /= 2; \
    bit--; \
    if ((*name & (1 << bit)) == 0) { \
      f1; \
      ChanIn(d_in[bit], (r), (s)); \
    } \
    else { \
      f2; \
      ChanOut(d_out[bit], (r), (s)); \
    } \
    num_processors *= 2; \
    bit++; \
  }
/*}}}  */

/*{{{  PAR(f1, r1, s1, f2, r2, s2)*/
#define PAR(f1, r1, s1, f2, r2, s2) \
  { \
    num_processors /= 2; \
    bit--; \
    if ((*name & (1 << bit)) == 0) { \
      f1; \
      ChanIn(d_in[bit], (r2), (s2)); \
      ChanOut(d_out[bit], (r1), (s1)); \
    } \
    else { \
      f2; \
      ChanOut(d_out[bit], (r2), (s2)); \
      ChanIn(d_in[bit], (r1), (s1)); \
    } \
    num_processors *= 2; \
    bit++; \
  }
/*}}}  */

typedef enum {FALSE = 0, TRUE = 1} BOOLEAN;

int node_level(int name);

int dim; /* current dimension */
Channel **d_in = NULL;                     /* Input Communication Channels */
Channel **d_out = NULL;                    /* Output Communication Channels */
int level;                  /* Processor level in the HyperCube */
int num_processors;     /* # of Processors in the group */
int *name;            /* Processor identifier */
int bit;
char OUTPUT[BUFFERSIZE];

/*{{{  Channel ** CreateChannels (Channel *OldChannels[], int ChannelsSize)*/
Channel ** CreateChannels (Channel *OldChannels[], int ChannelsSize)
{
    Channel **NewChannels = NULL;

    if ((NewChannels = malloc((ChannelsSize + 1) * sizeof(Channel *))) == NULL)
        abort();
    else
    {
        int ChannelsCount = 0;

        while (ChannelsCount++ < ChannelsSize)
            NewChannels[ChannelsCount - 1] = OldChannels[ChannelsCount - 1];

        NewChannels[ChannelsSize] = NULL;
    }
    return(NewChannels);
}
/*}}}  */

/*{{{  int node_level(int name)*/
int node_level(int name) {
  register int lev = 0, max_name = 0;

  while (max_name < name) {
    lev++;
    max_name = (max_name << 1) + 1;
  }
  return lev;
}
/*}}}  */

#ifdef ROOT
  #include "rsup.c"
#else
  #include "nsup.c"
#endif
